@@ -4,5 +4,5 @@
 #
 # Miscellaneous cli-related constants.
 #
-DEFAULT_CONTENT_FOLDER = 'downloaded'
+DEFAULT_CONTENT_FOLDER = "downloaded"
 DEFAULT_PREBUFFER_BYTES = 3_000_000  # 3mb
