@@ -19,40 +19,27 @@ def mediacloud_search_action(cli_args):
     client = MediacloudAPIClient(cli_args.token)
 
     kwargs = {
-        'collections': cli_args.collections,
-        'medias': cli_args.medias,
-        'publish_day': cli_args.publish_day,
-        'publish_month': cli_args.publish_month,
-        'publish_year': cli_args.publish_year,
-        'filter_query': cli_args.filter_query
+        "collections": cli_args.collections,
+        "medias": cli_args.medias,
+        "publish_day": cli_args.publish_day,
+        "publish_month": cli_args.publish_month,
+        "publish_year": cli_args.publish_year,
+        "filter_query": cli_args.filter_query,
     }
 
-    loading_bar = LoadingBar(
-        'Searching stories',
-        unit='story',
-        unit_plural='stories'
-    )
+    loading_bar = LoadingBar("Searching stories", unit="story", unit_plural="stories")
 
     try:
         if not cli_args.skip_count:
-            count = client.count(
-                cli_args.query,
-                **kwargs
-            )
+            count = client.count(cli_args.query, **kwargs)
 
             loading_bar.update_total(count)
 
-        iterator = client.search(
-            cli_args.query,
-            **kwargs
-        )
+        iterator = client.search(cli_args.query, **kwargs)
 
         for story in iterator:
             writer.writerow(story.as_csv_row())
             loading_bar.update()
 
     except MediacloudServerError as e:
-        loading_bar.die([
-            'Aborted due to a mediacloud server error:',
-            e.server_error
-        ])
+        loading_bar.die(["Aborted due to a mediacloud server error:", e.server_error])
