@@ -109,3 +109,18 @@ def format_reply(item, video_id=None):
     )
 
     return row
+
+
+def format_playlist_item_snippet(item):
+    snippet = item["snippet"]
+
+    row = YouTubeVideoSnippet(
+        snippet["resourceId"]["videoId"],
+        snippet["publishedAt"],
+        snippet["channelId"],
+        snippet["title"],
+        snippet["description"],
+        snippet["channelTitle"],
+    )
+
+    return row
